@@ -6,6 +6,7 @@ use strict;
 require 5.00503;
 
 use Carp qw(croak confess carp);
+use Data::Dumper;
 use IPC::Semaphore;
 use IPC::Shareable::SharedMem;
 use IPC::SysV qw(
@@ -15,10 +16,12 @@ use IPC::SysV qw(
     IPC_NOWAIT
     SEM_UNDO
 );
-use Storable 0.6 qw(freeze thaw);
+use JSON qw(-convert_blessed_universally);
 use Scalar::Util;
+use String::CRC32;
+use Storable 0.6 qw(freeze thaw);
 
-our $VERSION = '1.04';
+our $VERSION = '1.05';
 
 $SIG{CHLD} = 'IGNORE';
 
@@ -32,6 +35,8 @@ use constant {
     SHM_BUFSIZ  =>  65536,
     SEM_MARKER  =>  0,
     SHM_EXISTS  =>  1,
+
+    SHMMAX_BYTES    => 1073741824, # 1 GB
 };
 
 require Exporter;
@@ -80,14 +85,16 @@ my %semop_args = (
 );
 
 my %default_options = (
-    key       => IPC_PRIVATE,
-    create    => 0,
-    exclusive => 0,
-    destroy   => 0,
-    graceful  => 0,
-    warn      => 0,
-    mode      => 0666,
-    size      => SHM_BUFSIZ,
+    key        => IPC_PRIVATE,
+    create     => 0,
+    exclusive  => 0,
+    destroy    => 0,
+    mode       => 0666,
+    size       => SHM_BUFSIZ,
+    limit      => 1,
+    graceful   => 0,
+    warn       => 0,
+    serializer => 'storable',
 );
 
 my %global_register;
@@ -114,23 +121,21 @@ sub STORE {
 
     $global_register{$sid} ||= $knot;
 
-    $knot->{_data} = _thaw($knot->seg) unless ($knot->{_lock});
+    $knot->{_data} = $knot->_decode($knot->seg) unless ($knot->{_lock});
 
     if ($knot->{_type} eq 'HASH') {
-        my $key = shift;
-        my $val = shift;
-        _mg_tie($knot, $val) if _need_tie($val);
+        my ($key, $val) = @_;
+        _mg_tie($knot, $val, $key) if $knot->_need_tie($val, $key);
         $knot->{_data}{$key} = $val;
     }
     elsif ($knot->{_type} eq 'ARRAY') {
-        my $i   = shift;
-        my $val = shift;
-        _mg_tie($knot, $val) if _need_tie($val);
-        $knot->{_data}->[$i] = $val;
+        my ($i, $val) = @_;
+        _mg_tie($knot, $val, $i) if $knot->_need_tie($val, $i);
+        $knot->{_data}[$i] = $val;
     }
     elsif ($knot->{_type} eq 'SCALAR') {
-        my $val = shift;
-        _mg_tie($knot, $val) if _need_tie($val);
+        my ($val) = @_;
+        _mg_tie($knot, $val) if $knot->_need_tie($val);
         $knot->{_data} = \$val;
     }
     else {
@@ -140,7 +145,7 @@ sub STORE {
     if ($knot->{_lock} & LOCK_EX) {
         $knot->{_was_changed} = 1;
     } else {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!\n";
         }
     }
@@ -158,7 +163,7 @@ sub FETCH {
         $knot->{_iterating} = 0; # In case we break out
         $data = $knot->{_data};
     } else {
-        $data = _thaw($knot->seg);
+        $data = $knot->_decode($knot->seg);
         $knot->{_data} = $data;
     }
 
@@ -193,7 +198,7 @@ sub FETCH {
 
     if (my $inner = _is_kid($val)) {
         my $s = $inner->seg;
-        $inner->{_data} = _thaw($s);
+        $inner->{_data} = $knot->_decode($s);
     }
     return $val;
 
@@ -215,7 +220,7 @@ sub CLEAR {
     if ($knot->{_lock} & LOCK_EX) {
         $knot->{_was_changed} = 1;
     } else {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!";
         }
     }
@@ -224,12 +229,12 @@ sub DELETE {
     my $knot = shift;
     my $key  = shift;
 
-    $knot->{_data} = _thaw($knot->seg) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg) unless $knot->{_lock};
     my $val = delete $knot->{_data}->{$key};
     if ($knot->{_lock} & LOCK_EX) {
         $knot->{_was_changed} = 1;
     } else {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!";
         }
     }
@@ -240,14 +245,14 @@ sub EXISTS {
     my $knot = shift;
     my $key  = shift;
 
-    $knot->{_data} = _thaw($knot->seg) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg) unless $knot->{_lock};
     return exists $knot->{_data}->{$key};
 }
 sub FIRSTKEY {
     my $knot = shift;
 
     $knot->{_iterating} = 1;
-    $knot->{_data} = _thaw($knot->seg) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg) unless $knot->{_lock};
     my $reset = keys %{$knot->{_data}};
     my $first = each %{$knot->{_data}};
     return $first;
@@ -272,13 +277,13 @@ sub PUSH {
     my $knot = shift;
 
     $global_register{$knot->seg->id} ||= $knot;
-    $knot->{_data} = _thaw($knot->seg, $knot->{_data}) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg, $knot->{_data}) unless $knot->{_lock};
 
     push @{$knot->{_data}}, @_;
     if ($knot->{_lock} & LOCK_EX) {
         $knot->{_was_changed} = 1;
     } else {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!";
         };
     }
@@ -286,13 +291,13 @@ sub PUSH {
 sub POP {
     my $knot = shift;
 
-    $knot->{_data} = _thaw($knot->seg, $knot->{_data}) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg, $knot->{_data}) unless $knot->{_lock};
 
     my $val = pop @{$knot->{_data}};
     if ($knot->{_lock} & LOCK_EX) {
         $knot->{_was_changed} = 1;
     } else {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!";
         }
     }
@@ -301,12 +306,12 @@ sub POP {
 sub SHIFT {
     my $knot = shift;
 
-    $knot->{_data} = _thaw($knot->seg, $knot->{_data}) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg, $knot->{_data}) unless $knot->{_lock};
     my $val = shift @{$knot->{_data}};
     if ($knot->{_lock} & LOCK_EX) {
         $knot->{_was_changed} = 1;
     } else {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!";
         }
     }
@@ -315,12 +320,12 @@ sub SHIFT {
 sub UNSHIFT {
     my $knot = shift;
 
-    $knot->{_data} = _thaw($knot->seg, $knot->{_data}) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg, $knot->{_data}) unless $knot->{_lock};
     my $val = unshift @{$knot->{_data}}, @_;
     if ($knot->{_lock} & LOCK_EX) {
         $knot->{_was_changed} = 1;
     } else {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!";
         }
     }
@@ -329,12 +334,12 @@ sub UNSHIFT {
 sub SPLICE {
     my($knot, $off, $n, @av) = @_;
 
-    $knot->{_data} = _thaw($knot->seg, $knot->{_data}) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg, $knot->{_data}) unless $knot->{_lock};
     my @val = splice @{$knot->{_data}}, $off, $n, @av;
     if ($knot->{_lock} & LOCK_EX) {
         $knot->{_was_changed} = 1;
     } else {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!";
         }
     }
@@ -343,19 +348,19 @@ sub SPLICE {
 sub FETCHSIZE {
     my $knot = shift;
 
-    $knot->{_data} = _thaw($knot->seg) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg) unless $knot->{_lock};
     return scalar(@{$knot->{_data}});
 }
 sub STORESIZE {
     my $knot = shift;
     my $n    = shift;
 
-    $knot->{_data} = _thaw($knot->seg) unless $knot->{_lock};
+    $knot->{_data} = $knot->_decode($knot->seg) unless $knot->{_lock};
     $#{$knot->{_data}} = $n - 1;
     if ($knot->{_lock} & LOCK_EX) {
         $knot->{_was_changed} = 1;
     } else {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!";
         }
     }
@@ -371,6 +376,23 @@ sub process_register {
     return \%process_register;
 }
 
+sub attributes {
+    my ($knot, $attr) = @_;
+
+    my $attrs = $knot->{attributes};
+
+    if (defined $attr) {
+        return $knot->{attributes}{$attr};
+    }
+    else {
+        return $knot->{attributes};
+    }
+}
+sub ipcs {
+    my $count = `ipcs -m | wc -l`;
+    chomp $count;
+    return $count;
+}
 sub spawn {
     my ($knot, %opts) = @_;
 
@@ -443,7 +465,7 @@ sub lock {
     my $return_val = $sem->op(@{ $semop_args{$flags} });
     if ($return_val) {
         $knot->{_lock} = $flags;
-        $knot->{_data} = _thaw($knot->seg),
+        $knot->{_data} = $knot->_decode($knot->seg),
     }
     return $return_val;
 }
@@ -452,7 +474,7 @@ sub unlock {
 
     return 1 unless $knot->{_lock};
     if ($knot->{_was_changed}) {
-        if (! defined _freeze($knot->seg, $knot->{_data})){
+        if (! defined $knot->_encode($knot->seg, $knot->{_data})){
             croak "Could not write to shared memory: $!\n";
         }
         $knot->{_was_changed} = 0;
@@ -473,7 +495,7 @@ sub clean_up {
     my $class = shift;
 
     for my $s (values %process_register) {
-        next unless $s->{attributes}{owner} == $$;
+        next unless $s->attributes('owner') == $$;
         remove($s);
     }
 }
@@ -534,13 +556,76 @@ sub singleton {
 END {
     for my $s (values %process_register) {
         unlock($s);
-        next unless $s->{attributes}->{destroy};
-        next unless $s->{attributes}->{owner} == $$;
+        next unless $s->attributes('destroy');
+        next unless $s->attributes('owner') == $$;
         remove($s);
     }
 }
 
 # --- Private methods below
+
+sub _encode {
+    my ($knot, $seg, $data) = @_;
+
+    my $serializer = $knot->attributes('serializer');
+
+    if ($serializer eq 'storable') {
+        return _freeze($seg, $data);
+    }
+    elsif ($serializer eq 'json'){
+        return _encode_json($seg, $data);
+    }
+
+    return undef;
+}
+sub _decode {
+    my ($knot, $seg) = @_;
+
+    my $serializer = $knot->attributes('serializer');
+
+    if ($serializer eq 'storable') {
+        return _thaw($seg);
+    }
+    elsif ($serializer eq 'json'){
+        return _decode_json($seg);
+    }
+
+    return undef;
+}
+sub _encode_json {
+    my $seg  = shift;
+    my $data = shift;
+
+    my $json = encode_json $data;
+
+    if (length($json) > $seg->size) {
+        croak "Length of shared data exceeds shared segment size";
+    }
+    $seg->shmwrite($json);
+}
+sub _decode_json {
+    my $seg = shift;
+
+    my $json = $seg->shmread;
+
+    return if ! $json;
+
+    # Remove \x{0} after end of string (broke JSON)
+
+    $json =~ s/\x00+//;
+
+#    my $tag = substr $json, 0, 14, '';
+
+#    if ($tag eq 'IPC::Shareable') {
+        my $data = decode_json $json;
+        if (! defined($data)){
+            croak "Munged shared memory segment (size exceeded?)";
+        }
+        return $data;
+#    } else {
+#        return;
+#    }
+}
 sub _freeze {
     my $seg  = shift;
     my $water = shift;
@@ -574,24 +659,39 @@ sub _thaw {
     }
 }
 sub _tie {
-    my $type  = shift;
-    my $class = shift;
-    my $opts  = _parse_args(@_);
+    my ($type, $class, $key_str, $opts);
 
-    my $key      = _shm_key($opts);
-    my $flags    = _shm_flags($opts);
-    my $shm_size = $opts->{size};
+    if (scalar @_ == 4) {
+        ($type, $class, $key_str, $opts) = @_;
+        $opts->{key} = $key_str;
+    }
+    else {
+        ($type, $class, $opts) = @_;
+    }
+
+    $opts  = _parse_args($opts);
+
+    my $knot = bless { attributes => $opts }, $class;
+
+    my $key      = $knot->_shm_key;
+    my $flags    = $knot->_shm_flags;
+    my $shm_size = $knot->attributes('size');
+
+    if ($knot->attributes('limit') && $shm_size > SHMMAX_BYTES) {
+        croak
+            "Shared memory segment size '$shm_size' is larger than max size of " .              SHMMAX_BYTES;
+    }
 
     my $seg;
 
-    if ($opts->{graceful}) {
+    if ($knot->attributes('graceful')) {
         my $exclusive = eval {
             $seg = IPC::Shareable::SharedMem->new($key, $shm_size, $flags);
             1;
         };
 
         if (! defined $exclusive) {
-            if ($opts->{warn}) {
+            if ($knot->attributes('warn')) {
                 warn "Process ID $$ exited due to exclusive shared memory collision\n";
             }
             exit(0);
@@ -602,18 +702,28 @@ sub _tie {
     }
 
     if (! defined $seg) {
-        if (! $opts->{create}) {
+        if ($! =~ /Cannot allocate memory/) {
+            croak "\nERROR: Could not create shared memory segment: $!\n\n" .
+                  "Are you using too large a size?";
+        }
+
+        if ($! =~ /No space left on device/) {
+            croak "\nERROR: Could not create shared memory segment: $!\n\n" .
+                "Are you spawning too many segments in a loop?";
+        }
+
+        if (! $knot->attributes('create')) {
             confess "ERROR: Could not acquire shared memory segment... 'create' ".
                   "option is not set, and the segment hasn't been created " .
                   "yet:\n\n $!";
         }
-        elsif ($opts->{create} && $opts->{exclusive}){
+        elsif ($knot->attributes('create') && $knot->attributes('exclusive')){
             croak "ERROR: Could not create shared memory segment. 'create' " .
                   "and 'exclusive' are set. Does the segment already exist? " .
                   "\n\n$!";
         }
         else {
-            croak "ERROR: Could not create shared memory segment.\n\n$!n";
+            croak "ERROR: Could not create shared memory segment.\n\n$!";
         }
     }
 
@@ -626,8 +736,8 @@ sub _tie {
         croak "Could not obtain semaphore set lock: $!\n";
     }
 
-    my $knot = {
-        attributes   => $opts,
+    %$knot = (
+        %$knot,
         _iterating   => 0,
         _key         => $key,
         _lock        => 0,
@@ -635,13 +745,13 @@ sub _tie {
         _sem         => $sem,
         _type        => $type,
         _was_changed => 0,
-    };
+    );
 
     $knot->{_data} = _thaw($seg);
 
     if ($sem->getval(SEM_MARKER) != SHM_EXISTS) {
-        $global_register{$knot->{_shm}->id} ||= $knot;
-        $process_register{$knot->{_shm}->id} ||= $knot;
+        $global_register{$knot->seg->id} ||= $knot;
+        $process_register{$knot->seg->id} ||= $knot;
         if (! $sem->setval(SEM_MARKER, SHM_EXISTS)){
             croak "Couldn't set semaphore during object creation: $!";
         }
@@ -649,20 +759,13 @@ sub _tie {
 
     $sem->op(@{ $semop_args{(LOCK_SH|LOCK_UN)} });
 
-    return bless $knot, $class;
+    return $knot;
 }
 sub _parse_args {
-    my($proto, $opts) = @_;
+    my ($opts) = @_;
 
-    $proto = defined $proto ? $proto :  0;
     $opts  = defined $opts  ? $opts  : { %default_options };
 
-    if (ref $proto eq 'HASH') {
-        $opts = $proto;
-    }
-    else {
-        $opts->{key} = $proto;
-    }
     for my $k (keys %default_options) {
         if (not defined $opts->{$k}) {
             $opts->{$k} = $default_options{$k};
@@ -681,50 +784,63 @@ sub _parse_args {
     return $opts;
 }
 sub _shm_key {
-    my $hv = shift;
-    my $val = ($hv->{key} or '');
+    my ($knot) = @_;
 
-    if ($val eq '') {
-        return IPC_PRIVATE;
+    my $key_str = ($knot->attributes('key') or '');
+    my $key;
+
+    if ($key_str eq '') {
+        $key = IPC_PRIVATE;
     }
-    elsif ($val =~ /^\d+$/) {
-        return $val;
+    elsif ($key_str =~ /^\d+$/) {
+        $key = $key_str;
     }
     else {
-        # XXX This only uses the first four characters
-        $val = pack   'A4', $val;
-        $val = unpack 'i', $val;
-        return $val;
+        $key = crc32($key_str);
     }
+
+    $used_ids{$key}++;
+
+    return $key;
+}
+sub _shm_key_rand {
+    my $key;
+
+    do {
+        $key = int(rand(1_000_000));
+    } while ($used_ids{$key});
+
+    $used_ids{$key}++;
+
+    return $key;
 }
 sub _shm_flags {
     # --- Parses the anonymous hash passed to constructors; returns a list
     # --- of args suitable for passing to shmget
-    my $hv = shift;
+    my ($knot) = @_;
+
     my $flags = 0;
 
-    $flags |= IPC_CREAT if $hv->{create};
-    $flags |= IPC_EXCL  if $hv->{exclusive};
-    $flags |= ($hv->{mode} or 0666);
+    $flags |= IPC_CREAT if $knot->attributes('create');
+    $flags |= IPC_EXCL  if $knot->attributes('exclusive');;
+    $flags |= ($knot->attributes('mode') or 0666);
 
     return $flags;
 }
 sub _mg_tie {
-    my $parent = shift;
-    my $val = shift;
+    my ($parent, $val, $identifier) = @_;
 
-    # XXX How to generate a unique id ?
     my $key;
+
     if ($parent->{_key} == IPC_PRIVATE) {
         $key = IPC_PRIVATE;
-    } else {
-        do {
-            $key = int(rand(1_000_000));
-        } while ($used_ids{$key});
-        $used_ids{$key}++;
     }
+    else {
+        $key = _shm_key_rand();
+    }
+
     my %opts = (
-        %{$parent->{attributes}},
+        %{ $parent->attributes },
         key       => $key,
         exclusive => 1,
         create    => 1,
@@ -734,24 +850,31 @@ sub _mg_tie {
     # XXX I wish I didn't have to take a copy of data here and copy it back in
     # XXX Also, have to peek inside potential objects to see their implementation
     my $child;
-    my $type = Scalar::Util::reftype( $val ) || '';
+    my $type = Scalar::Util::reftype($val) || '';
 
     if ($type eq "HASH") {
         my %copy = %$val;
         $child = tie %$val, 'IPC::Shareable', $key, { %opts };
         croak "Could not create inner tie" if ! $child;
+
+        _reset_segment($parent, $identifier) if $opts{tidy};
+
         %$val = %copy;
     }
     elsif ($type eq "ARRAY") {
         my @copy = @$val;
         $child = tie @$val, 'IPC::Shareable', $key, { %opts };
         croak "Could not create inner tie" if ! $child;
+
+        _reset_segment($parent, $identifier) if $opts{tidy};
+
         @$val = @copy;
     }
     elsif ($type eq "SCALAR") {
         my $copy = $$val;
         $child = tie $$val, 'IPC::Shareable', $key, { %opts };
         croak "Could not create inner tie" if ! $child;
+
         $$val = $copy;
     }
     else {
@@ -785,22 +908,42 @@ sub _is_kid {
     return;
 }
 sub _need_tie {
-    my $val = shift;
+    my ($knot, $val, $identifier) = @_;
 
-    my $type = Scalar::Util::reftype( $val );
-    return unless $type;
+    my $type = Scalar::Util::reftype($val);
+    return 0 if ! $type;
+
+    my $need_tie;
 
     if ($type eq "HASH") {
-        return !(tied %$val);
+        $need_tie = !(tied %$val);
     }
     elsif ($type eq "ARRAY") {
-        return !(tied @$val);
+        $need_tie = !(tied @$val);
     }
     elsif ($type eq "SCALAR") {
-        return !(tied $$val);
+        $need_tie = !(tied $$val);
     }
 
-    return;
+    return $need_tie ? 1 : 0;
+}
+sub _reset_segment {
+    my ($parent, $id) = @_;
+
+    my $parent_type = Scalar::Util::reftype($parent->{_data}) || '';
+
+    if ($parent_type eq 'HASH') {
+        my $data = $parent->{_data};
+        if (exists $data->{$id} && keys %{ $data->{$id} } && tied %{ $data->{$id} }) {
+            (tied %{ $parent->{_data}{$id} })->remove;
+        }
+    }
+    elsif ($parent_type eq 'ARRAY') {
+        my $data = $parent->{_data};
+        if (exists $data->[$id] && tied @{ $data->[$id] }) {
+            (tied @{ $parent->{_data}[$id] })->remove;
+        }
+    }
 }
 
 sub _trace {
@@ -813,7 +956,7 @@ sub _trace {
         my $obj;
         if (ref eq 'IPC::Shareable') {
             '        ' . "\$_[$i] = $_: shmid: $_->{_shm}->{_id}; " .
-                Data::Dumper->Dump([ $_->{attributes} ], [ 'opts' ]);
+                Data::Dumper->Dump([ $_->attributes ], [ 'opts' ]);
         } else {
             '        ' . Data::Dumper->Dump( [ $_ ] => [ "\_[$i]" ]);
         }
@@ -829,7 +972,7 @@ sub _debug {
         my $obj;
         if (ref eq 'IPC::Shareable') {
             '        ' . "$_: shmid: $_->{_shm}->{_id}; " .
-                Data::Dumper->Dump([ $_->{attributes} ], [ 'opts' ]);
+                Data::Dumper->Dump([ $_->attributes ], [ 'opts' ]);
         }
         else {
             '        ' . Data::Dumper::Dumper($_);
@@ -876,7 +1019,7 @@ IPC::Shareable - Use shared memory backed variables across processes
 
     # Ensure only one instance of a script can be run at any time
 
-    IPC::Shareable->singleton('GLUE');
+    IPC::Shareable->singleton('UNIQUE SCRIPT LOCK STRING');
 
 =head1 DESCRIPTION
 
@@ -884,39 +1027,38 @@ IPC::Shareable allows you to tie a variable to shared memory making it
 easy to share the contents of that variable with other Perl processes and
 scripts.
 
-Scalars, arrays, and hashes and even objects can be tied. The variable being
+Scalars, arrays, hashes and even objects can be tied. The variable being
 tied may contain arbitrarily complex data structures - including references to
 arrays, hashes of hashes, etc.
 
 The association between variables in distinct processes is provided by
-GLUE (aka "key").  This is an integer number or 4 character string[1] that
-serves as a common identifier for data across process space.  Hence the
-statement
+GLUE (aka "key").  This is any arbitrary string or integer that serves as a
+common identifier for data across process space.  Hence the statement:
 
-    tie my $scalar, 'IPC::Shareable', { key => 'GLUE', create => 1 };
+    tie my $scalar, 'IPC::Shareable', { key => 'GLUE STRING', create => 1 };
 
 ...in program one and the statement
 
-    tie my $variable, 'IPC::Shareable', { key => 'GLUE' };
+    tie my $variable, 'IPC::Shareable', { key => 'GLUE STRING' };
 
 ...in program two will create and bind C<$scalar> the shared memory in program
 one and bind it to C<$variable> in program two.
 
 There is no pre-set limit to the number of processes that can bind to
 data; nor is there a pre-set limit to the complexity of the underlying
-data of the tied variables[2].  The amount of data that can be shared
+data of the tied variables.  The amount of data that can be shared
 within a single bound variable is limited by the system's maximum size
 for a shared memory segment (the exact value is system-dependent).
 
 The bound data structures are all linearized (using Raphael Manfredi's
-L<Storable> module) before being slurped into shared memory.  Upon
-retrieval, the original format of the data structure is recovered.
+L<Storable> module or optionally L<JSON>) before being slurped into shared
+memory.  Upon retrieval, the original format of the data structure is recovered.
 Semaphore flags can be used for locking data between competing processes.
 
 =head1 OPTIONS
 
-Options are specified by passing a reference to a hash as the fourth
-argument to the C<tie()> function that enchants a variable.
+Options are specified by passing a reference to a hash as the third argument to
+the C<tie()> function that enchants a variable.
 
 The following fields are recognized in the options hash:
 
@@ -932,10 +1074,10 @@ Default: B<IPC_PRIVATE>
 
 =head2 create
 
-B<create> is used to control whether calls to C<tie()> create new shared
-memory segments or not.  If B<create> is set to a true value,
+B<create> is used to control whether the process creates a new shared
+memory segment or not.  If B<create> is set to a true value,
 L<IPC::Shareable> will create a new binding associated with GLUE as
-needed.  If B<create> is false, C<IPC::Shareable> will not attempt to
+needed.  If B<create> is false, L<IPC::Shareable> will not attempt to
 create a new shared memory segment associated with GLUE.  In this
 case, a shared memory segment associated with GLUE must already exist
 or we'll C<croak()>.
@@ -949,14 +1091,14 @@ binding associated with GLUE already exists.  If set to a false value, calls to
 C<tie()> will succeed even if a shared memory segment associated with GLUE
 already exists.
 
-See L</graceful> for a silent, non-exception exit if a second process
-attempts to obtain an in-use C<exclusive> segment.
+See L</graceful> for a silent, non-exception exit if a second process attempts
+to obtain an in-use C<exclusive> segment.
 
 Default: B<false>
 
 =head2 graceful
 
-If B<exclusive> is set to a true value, we normally C<die()> if a second
+If B<exclusive> is set to a true value, we normally C<croak()> if a second
 process attempts to obtain the same shared memory segment. Set B<graceful>
 to true and we'll C<exit> silently and gracefully. This option does nothing
 if C<exclusive> isn't set.
@@ -965,6 +1107,13 @@ Useful for ensuring only a single process is running at a time.
 
 Default: B<false>
 
+=head2 warn
+
+When set to a true value, B<graceful> will output a warning if there are
+process collisions.
+
+Default: B<false>
+
 =head2 mode
 
 The B<mode> argument is an octal number specifying the access
@@ -975,10 +1124,31 @@ process creating the shared variable, etc.
 
 Default: B<0666> (world read and writeable)
 
+=head2 size
+
+This field may be used to specify the size of the shared memory segment
+allocated.
+
+The maximum size we allow by default is ~1GB. See the L</limit> option to
+override this default.
+
+Default: C<IPC::Shareable::SHM_BUFSIZ()> (ie. B<65536>)
+
+=head2 limit
+
+This field will allow you to set a segment size larger than the default maximum
+which is 1,073,741,824 bytes (approximately 1 GB). If set, we will
+C<croak()> if a size specified is larger than the maximum. If it's set to a
+false value, we'll C<croak()> if you send in a size larger than the total
+system RAM.
+
+Default: B<true>
+
 =head2 destroy
 
 If set to a true value, the shared memory segment underlying the data
-binding will be removed when the process calling C<tie()> exits (gracefully)[3].
+binding will be removed when the process that initialized the shared memory
+segment exits (gracefully)[1].
 
 Only those memory segments that were created by the current process will be
 removed.
@@ -989,25 +1159,39 @@ is a finite resource and should be released if it is not needed.
 
 Default: B<false>
 
-=head2 size
+=head2 tidy
 
-This field may be used to specify the size of the shared memory segment
-allocated.
+For long running processes, set this to a true value to clean up unneeded
+segments from nested data structures. Comes with a slight performance hit.
 
-Default: C<IPC::Shareable::SHM_BUFSIZ()> (ie. B<65536>)
+Default: B<false>
+
+=head2 serializer
+
+By default, we use L<Storable> as the data serializer when writing to or
+reading from the shared memory segments we create. For cross-platform and
+cross-language purposes, you can optionally use L<JSON> for this task.
+
+Send in either C<json> or C<storable> as the value to use the respective
+serializer.
+
+Default: B<storable>
 
 =head2 Default Option Values
 
 Default values for options are:
 
-     key       => IPC_PRIVATE,
-     create    => 0,
-     exclusive => 0,
-     graceful  => 0,
-     warn      => 0,
-     destroy   => 0,
-     mode      => 0,
-     size      => IPC::Shareable::SHM_BUFSIZ(),
+    key         => IPC_PRIVATE,
+    create      => 0,
+    exclusive   => 0,
+    mode        => 0,
+    size        => IPC::Shareable::SHM_BUFSIZ(),
+    limit       => 1,
+    destroy     => 0,
+    graceful    => 0,
+    warn        => 0,
+    tidy        => 0,
+    serializer  => 'storable',
 
 =head1 METHODS
 
@@ -1024,11 +1208,18 @@ Mandatory, String: The key/glue that identifies the shared memory segment.
 
     $warn
 
-Bool, Optional: Send in a true value to have subsequent processes throw a
+Optional, Bool: Send in a true value to have subsequent processes throw a
 warning that there's been a shared memory violation and that it will exit.
 
 Default: B<false>
 
+=head2 ipcs
+
+Returns the number of instantiated shared memory segments that currently exist
+on the system.
+
+Return: Integer
+
 =head2 spawn(%opts)
 
 Spawns a forked process running in the background that holds the shared memory
@@ -1051,12 +1242,12 @@ Example:
 
     use IPC::Shareable;
 
-    # following line sets things up and returns
+    # The following line sets things up and returns
 
-    IPC::Shareable->spawn(key => 'GLUE');
+    IPC::Shareable->spawn(key => 'GLUE STRING');
 
 Now, either within the same script, or any other script on the system, your
-data will be available at the key/glue C<GLUE>. Call
+data will be available at the key/glue C<GLUE STRING>. Call
 L<unspawn()|/unspawn($key, $destroy)> to remove it.
 
 =head2 unspawn($key, $destroy)
@@ -1147,6 +1338,20 @@ memory segment object currently in use.
 Called on either the tied variable or the tie object, returns the semaphore
 object related to the memory segment currently in use.
 
+=head2 attributes
+
+Retrieves the list of attributes that drive the L<IPC::Shareable> object.
+
+Parameters:
+
+    $attribute
+
+Optional, String: The name of the attribute. If sent in, we'll return the value
+of this specific attribute. Returns C<undef> if the attribute isn't found.
+
+Returns: A hash reference of all attributes if C<$attributes> isn't sent in, the
+value of the specific attribute if it is.
+
 =head2 global_register
 
 Returns a hash reference of hashes of all in-use shared memory segments across
@@ -1288,6 +1493,10 @@ instance of C<IPC::Shareable> on success, and C<undef> otherwise.
 
 Benjamin Sugars <bsugars@canoe.ca>
 
+=head1 MAINTAINED BY
+
+Steve Bertrand <steveb@cpan.org>
+
 =head1 NOTES
 
 =head2 Footnotes from the above sections
@@ -1296,20 +1505,6 @@ Benjamin Sugars <bsugars@canoe.ca>
 
 =item 1
 
-If GLUE is longer than 4 characters, only the 4 most significant
-characters are used.  These characters are turned into integers by
-unpack()ing them.  If GLUE is less than 4 characters, it is space
-padded.
-
-=item 2
-
-IPC::Shareable provides no pre-set limits, but the system does.
-Namely, there are limits on the number of shared memory segments that
-can be allocated and the total amount of memory usable by shared
-memory.
-
-=item 3
-
 If the process has been smoked by an untrapped signal, the binding
 will remain in shared memory.  If you're cautious, you might try
 
@@ -1346,9 +1541,8 @@ locking implemented with semaphores in C<IPC::Shareable> is that when a
 process dies, it automatically releases any locks.  This only happens
 with C<IPC::Shareable> if the process dies gracefully.
 
-The alternative
-is to attempt to account for every possible calamitous ending for your
-process (robust signal handling in Perl is a source of much debate,
+The alternative is to attempt to account for every possible calamitous ending
+for your process (robust signal handling in Perl is a source of much debate,
 though it usually works just fine) or to become familiar with your
 system's tools for removing shared memory and semaphores.  This
 concern should be balanced against the significant performance
@@ -1368,10 +1562,18 @@ Examples:
 
     ipcs -a
 
-    # List all memory segments along with each one's associated process ID
+    # List all memory segments and semaphores along with each one's associated process ID
 
     ipcs -ap
 
+    # List just the shared memory segments
+
+    ipcs -m
+
+    # List the details of an individual memory segment
+
+    ipcs -i 12345678
+
     # Remove *all* semaphores and memory segments
 
     ipcrm -a
